package cn.jlsxwkj.moudles.chat;

import cn.dev33.satoken.SaManager;
import cn.hutool.crypto.digest.MD5;
import cn.jlsxwkj.common.reader.ParagraphDocReader;
import cn.jlsxwkj.common.reader.ParagraphTextReader;
import cn.jlsxwkj.common.utils.Log;
import cn.jlsxwkj.common.utils.MergeDocuments;
import cn.jlsxwkj.moudles.chathistory.ChatHistoryService;
import jakarta.annotation.Resource;
import org.springframework.ai.chat.messages.AssistantMessage;
import org.springframework.ai.chat.messages.Message;
import org.springframework.ai.chat.messages.SystemMessage;
import org.springframework.ai.chat.messages.UserMessage;
import org.springframework.ai.chat.prompt.Prompt;
import org.springframework.ai.document.Document;
import org.springframework.ai.document.DocumentReader;
import org.springframework.ai.ollama.OllamaChatModel;
import org.springframework.ai.vectorstore.SearchRequest;
import org.springframework.ai.vectorstore.VectorStore;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;
import reactor.core.publisher.Flux;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Objects;
import java.util.stream.Collectors;


/**
 * @author zh
 */
@Service
public class ChatService {

	/**
	 * 向量
	 */
	@Resource
	private VectorStore vectorStore;
	/**
	 * 大模型
	 */
	@Resource
	private OllamaChatModel ollamaChatModel;
	/**
	 * 历史记录
	 */
	@Resource
	private ChatHistoryService chatHistoryService;
	/**
	 * 单条消息
	 */
	private static StringBuffer chatMessage;
	/**
	 * md5 校验
	 */
	private static final MD5 MD5 = cn.hutool.crypto.digest.MD5.create();
	/**
	 * 当前项目路径 + /ai_doc
	 */
	private static final String PATH = Objects.requireNonNull(ChatService.class.getClassLoader().getResource("")).getPath() + "\\ai_doc\\";
	/**
	 * 用户对话上下文
	 */
	private static final List<Message> LIST_MESSAGE = new ArrayList<>();

	/**
	 * 使用spring ai解析txt文档并保存至 pg
	 * @param file 保存的文件
	 * @return 保存状态
	 * @throws IOException 保存失败
	 */
	public String uploadDocument(MultipartFile file) throws IOException {
		//分割文件名为 文件名, 后缀
		String[] split = Objects.requireNonNull(file.getOriginalFilename()).split("\\.");
		//获取文件类型
		String fileType = split[split.length - 1].toLowerCase(Locale.ROOT);
		//文件夹地址
		File savePath = new File(PATH);
		//保存文件对象
		File saveFile = new File(PATH + ChatService.MD5.digestHex(file.getInputStream()) + "." + fileType);
		//文件保存地址
		String fileUrl = saveFile.toURI().toURL().toString();
		//判断文件夹是否存在
		if (!savePath.exists()) {
			if (!savePath.mkdirs()) {
				Log.warn(this.getClass(), "创建文件夹失败: " + PATH);
			}
		}
		//判断文件类型
		DocumentReader reader;
		switch (fileType) {
			case "txt" -> reader = new ParagraphTextReader(fileUrl, 5);
			case "doc", "docx" -> reader = new ParagraphDocReader(fileUrl, 5);
			default -> throw new IOException("暂不支持的文件类型: " + fileType);
		}
		// 判断文件是否存在
		if (!saveFile.exists()) {
			try	{
				file.transferTo(saveFile);
				vectorStore.add(reader.get());
				return "上传完成";
			} catch (Exception e){
				boolean delete = saveFile.delete();
				if (!delete) {
					throw new IOException("删除文件失败: " + fileUrl);
				}
			}
		}
		throw new IOException("未知错误");
	}

	/**
	 * 根据关键词搜索向量库
	 *
	 * @param keyword 	关键词
	 * @return 			文本内容
	 */
	public String search(String keyword) {
		return MergeDocuments.mergeDocuments(
				vectorStore.similaritySearch(SearchRequest.query(keyword).withSimilarityThreshold(0.5))
		).stream().map(Document::getContent).collect(Collectors.joining("\n"));
	}

	/**
	 * 问答流,根据输入内容回答
	 *
	 * @param message 	输入内容
	 * @return 			回答内容
	 */
	public Flux<String> chatStream(String message) {
		chatMessage = new StringBuffer();
		//查询获取文档信息
		String content = search(message);
		//封装prompt并调用大模型
		Prompt chatPrompt2String = getChatPrompt2String(message, content);
		return ollamaChatModel.stream(chatPrompt2String).map(
				response -> {
					String str = response.getResult().getOutput().getContent();
					System.out.print(str);
					chatMessage.append(str);
					return str.replace(" ", "  ");
				})
				.concatWithValues("<{完成}>").onErrorStop()
				.doOnComplete(() -> {
					System.out.println();
					LIST_MESSAGE.add(new AssistantMessage(chatMessage.toString()));
					saveDialog(message, chatMessage.toString());
				});

	}

	/**
	 * 问答,根据输入内容回答
	 *
	 * @param message 	输入内容
	 * @return 			回答内容
	 */
	public String chat(String message) {
		chatMessage = new StringBuffer();
		//查询获取文档信息
		String content = search(message);
		//获取消息
		String result = ollamaChatModel.call(getChatPrompt2String(message, content))
				.getResult().getOutput().getContent();
		chatMessage.append(result);
		Log.info(this.getClass(),"{} message : {}", "chat", result);
		LIST_MESSAGE.add(new AssistantMessage(chatMessage.toString()));
		saveDialog(message, chatMessage.toString());
		return result;
	}

	/**
	 * 获取prompt
	 *
	 * @param message 	提问内容
	 * @param context 	上下文
	 * @return 			提示词
	 */
	private Prompt getChatPrompt2String(String message, String context) {
		Log.info(this.getClass(),"{} message : {}", "user", message);
		Prompt prompt = new Prompt(LIST_MESSAGE);
		LIST_MESSAGE.add(new SystemMessage(context));
		LIST_MESSAGE.add(new UserMessage(message));
		return prompt;
	}

	/**
	 * 保存会话
	 */
	private void saveDialog(String userQ, String chatA) {
		String userId = SaManager.getStpLogic("").getLoginIdAsString();
		chatHistoryService.insert(userId, userQ, chatA);
	}
}
